#
import json


#
from urllib import parse, request


#
from oaiv.tools.utils import format_provider, format_w3, data_constructor, check_precision
from oaiv.tools.address import find_address


#
class InteractionFunctionality:

    def __init__(self, etherscan_api_key, ethereum_network, infura_project_id):

        self.network = ethereum_network
        self.etherscan_api_key = etherscan_api_key
        self.provider = format_provider(ethereum_network=ethereum_network, infura_project_id=infura_project_id)
        self.w3 = format_w3(provider=self.provider)

        self.etherscan = EtherscanInteraction(network=ethereum_network, etherscan_api_key=etherscan_api_key)
        self.infura = InfuraInteraction(w3=self.w3)

    def balance(self, **kwargs):

        return self.etherscan.balance(**kwargs)

    def get_transactions(self, **kwargs):

        return self.etherscan.get_transactions(**kwargs)

    def create_account(self):

        return self.infura.create_account()

    def make_transaction(self, **kwargs):

        return self.infura.make_transaction(**kwargs)


class EtherscanInteraction:

    def __init__(self, network, etherscan_api_key):
        self.network = network
        self.etherscan_api_key = etherscan_api_key

    def request(self, params):
        if self.network == 'mainnet':
            url = 'https://api.etherscan.io/api'
        elif self.network == 'goerli':
            url = 'https://api-goerli.etherscan.io/api'
        elif self.network == 'ropsten':
            url = 'https://api-ropsten.etherscan.io/api'
        else:
            raise KeyError("Invalid network name")
        query = parse.urlencode(params)
        url = '{0}?{1}'.format(url, query)
        with request.urlopen(url) as response:
            response_data = json.loads(response.read())

        return response_data

    def balance(self, address, currency, status='latest', contract_address=None):

        if status not in ['latest', 'pending']:
            raise KeyError("Invalid status value. Please provide 'latest' or 'pending' status value.")
        if (contract_address is None) and (currency != 'ETH'):
            contract_address = find_address(name=currency)

        if currency == 'ETH':
            params = {
                'module': 'account',
                'action': 'balance',
                'address': address,
                'tag': status,
                'apikey': self.etherscan_api_key,
            }
        else:
            params = {
                'module': 'account',
                'action': 'tokenbalance',
                'contractaddress': contract_address,
                'address': address,
                'tag': status,
                'apikey': self.etherscan_api_key,
            }

        response_data = self.request(params)

        # TODO: fix the Decimal issue
        if response_data['message'] == 'OK':
            if currency == 'ETH':
                precision = 18
            else:
                precision = check_precision(currency)
            result = int(response_data['result']) / (10 ** precision)
        else:
            raise Exception("{0}".format(response_data))

        return result

    def get_transactions(self, account, sort='desc'):

        params = {
            'module': 'account',
            'action': 'txlist',
            'address': account,
            'startblock': 0,  # check numbers
            'endblock': 99999999,  # TODO: check numbers
            # 'page': 1,
            # 'offset': 10,
            'sort': sort,
            'apikey': self.etherscan_api_key,
        }
        response_data = self.request(params)
        return response_data


# TODO: add mnemonic support (see the w3.eth.account docs)
# TODO: add importing & exporting features
class Actor:
    def __init__(self, w3, private_key=None, public_key=None, encryption=None):
        self.w3 = w3
        self.private_key = private_key
        if private_key:
            self.account = w3.eth.account.from_key(private_key)
        else:
            self.account = None
        self.public_key = public_key
        self.encryption = encryption

    @property
    def nonce(self):
        return self.w3.eth.get_transaction_count(self.address)

    @property
    def address(self):
        if self.account:
            return self.account.address
        else:
            return self.w3.toChecksumAddress(self.public_key)

    def sign_transaction(self, tx):
        if self.private_key:
            return self.account.sign_transaction(tx)
        else:
            raise Exception("You have to provide private_key to use this feature")


class InfuraInteraction:

    def __init__(self, w3):

        self.w3 = w3

    # TODO: add mnemonic support (see the w3.eth.account docs)
    def create_account(self):
        private_key = self.w3.eth.account.create().key
        actor = Actor(w3=self.w3, private_key=private_key)
        return actor

    def generate_transaction_data(self, sender, receiver, value=None, currency=None, gas=None):

        tx = {
            'from': sender.address,
            'to': receiver.address,
        }

        if value:
            if currency == 'ETH':
                tx['value'] = self.w3.toWei(value, 'ether')
            else:
                tx['data'] = data_constructor(receiver_address=receiver.address, amount=value, currency=currency)

        # TODO: improve gas calculations with pre-London and post-London versions
        if gas:
            tx['gas'] = gas
        else:
            tx['gas'] = self.w3.eth.estimate_gas(tx)

        tx['gasPrice'] = self.w3.eth.gasPrice
        tx['nonce'] = sender.nonce

        return tx

    def make_transaction(self, sender, receiver, value=None, currency=None, gas=None):

        tx = self.generate_transaction_data(sender=sender, receiver=receiver,
                                            value=value, currency=currency, gas=gas)

        signed_txn = sender.sign_transaction(tx)

        tx_id = self.w3.toHex(self.w3.eth.sendRawTransaction(signed_txn.rawTransaction))

        return tx_id



#
import json


#
from web3 import Web3
from web3.middleware import geth_poa_middleware


#


#
def check_precision(currency):

    if currency == 'USDC':
        precision = 6
    else:
        raise KeyError("Unknown contract name")

    """
    if network in ['mainnet', 'goerli', 'ropsten']:

        with open('./oaiv/tools/address/{0}/info.json'.format(network), 'r') as file:
            names = json.load(file)
        if currency in names.keys():
            precision = int(names[currency]['precision'])
        else:
            raise KeyError("Unknown currency name")

        return precision

    else:

        raise KeyError("Invalid network name")
    """

    return precision


def data_constructor(receiver_address, amount, currency):
    method = '0xa9059cbb'
    receiver = "0" * (64 - len(receiver_address[2:])) + receiver_address[2:]
    amount_precision = check_precision(currency=currency)
    amount = hex(int(amount * (10 ** amount_precision)))[2:]
    amount = "0" * (64 - len(amount)) + amount
    data = method + receiver + amount
    return data


def format_provider(ethereum_network, infura_project_id):

    provider = 'https://{0}.infura.io/v3/{1}'.format(
        ethereum_network,
        infura_project_id
    )

    return provider


def format_w3(provider):

    w3 = Web3(Web3.HTTPProvider(
        provider
    ))
    w3.middleware_onion.inject(geth_poa_middleware, layer=0)

    return w3



#
import json


#


#


#
def load_contract(name, w3):

    address = find_address(name=name)

    if name == 'QUOTER':
        abi = '[{"inputs":[{"internalType":"address","name":"_factory","type":"address"},{"internalType":"address","name":"_WETH9","type":"address"}],"stateMutability":"nonpayable","type":"constructor"},{"inputs":[],"name":"WETH9","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"factory","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes","name":"path","type":"bytes"},{"internalType":"uint256","name":"amountIn","type":"uint256"}],"name":"quoteExactInput","outputs":[{"internalType":"uint256","name":"amountOut","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"tokenIn","type":"address"},{"internalType":"address","name":"tokenOut","type":"address"},{"internalType":"uint24","name":"fee","type":"uint24"},{"internalType":"uint256","name":"amountIn","type":"uint256"},{"internalType":"uint160","name":"sqrtPriceLimitX96","type":"uint160"}],"name":"quoteExactInputSingle","outputs":[{"internalType":"uint256","name":"amountOut","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes","name":"path","type":"bytes"},{"internalType":"uint256","name":"amountOut","type":"uint256"}],"name":"quoteExactOutput","outputs":[{"internalType":"uint256","name":"amountIn","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"tokenIn","type":"address"},{"internalType":"address","name":"tokenOut","type":"address"},{"internalType":"uint24","name":"fee","type":"uint24"},{"internalType":"uint256","name":"amountOut","type":"uint256"},{"internalType":"uint160","name":"sqrtPriceLimitX96","type":"uint160"}],"name":"quoteExactOutputSingle","outputs":[{"internalType":"uint256","name":"amountIn","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"int256","name":"amount0Delta","type":"int256"},{"internalType":"int256","name":"amount1Delta","type":"int256"},{"internalType":"bytes","name":"path","type":"bytes"}],"name":"uniswapV3SwapCallback","outputs":[],"stateMutability":"view","type":"function"}]'
    elif name == 'ROUTER':
        abi = '[{"inputs":[{"internalType":"address","name":"_factory","type":"address"},{"internalType":"address","name":"_WETH9","type":"address"}],"stateMutability":"nonpayable","type":"constructor"},{"inputs":[],"name":"WETH9","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"components":[{"internalType":"bytes","name":"path","type":"bytes"},{"internalType":"address","name":"recipient","type":"address"},{"internalType":"uint256","name":"deadline","type":"uint256"},{"internalType":"uint256","name":"amountIn","type":"uint256"},{"internalType":"uint256","name":"amountOutMinimum","type":"uint256"}],"internalType":"struct ISwapRouter.ExactInputParams","name":"params","type":"tuple"}],"name":"exactInput","outputs":[{"internalType":"uint256","name":"amountOut","type":"uint256"}],"stateMutability":"payable","type":"function"},{"inputs":[{"components":[{"internalType":"address","name":"tokenIn","type":"address"},{"internalType":"address","name":"tokenOut","type":"address"},{"internalType":"uint24","name":"fee","type":"uint24"},{"internalType":"address","name":"recipient","type":"address"},{"internalType":"uint256","name":"deadline","type":"uint256"},{"internalType":"uint256","name":"amountIn","type":"uint256"},{"internalType":"uint256","name":"amountOutMinimum","type":"uint256"},{"internalType":"uint160","name":"sqrtPriceLimitX96","type":"uint160"}],"internalType":"struct ISwapRouter.ExactInputSingleParams","name":"params","type":"tuple"}],"name":"exactInputSingle","outputs":[{"internalType":"uint256","name":"amountOut","type":"uint256"}],"stateMutability":"payable","type":"function"},{"inputs":[{"components":[{"internalType":"bytes","name":"path","type":"bytes"},{"internalType":"address","name":"recipient","type":"address"},{"internalType":"uint256","name":"deadline","type":"uint256"},{"internalType":"uint256","name":"amountOut","type":"uint256"},{"internalType":"uint256","name":"amountInMaximum","type":"uint256"}],"internalType":"struct ISwapRouter.ExactOutputParams","name":"params","type":"tuple"}],"name":"exactOutput","outputs":[{"internalType":"uint256","name":"amountIn","type":"uint256"}],"stateMutability":"payable","type":"function"},{"inputs":[{"components":[{"internalType":"address","name":"tokenIn","type":"address"},{"internalType":"address","name":"tokenOut","type":"address"},{"internalType":"uint24","name":"fee","type":"uint24"},{"internalType":"address","name":"recipient","type":"address"},{"internalType":"uint256","name":"deadline","type":"uint256"},{"internalType":"uint256","name":"amountOut","type":"uint256"},{"internalType":"uint256","name":"amountInMaximum","type":"uint256"},{"internalType":"uint160","name":"sqrtPriceLimitX96","type":"uint160"}],"internalType":"struct ISwapRouter.ExactOutputSingleParams","name":"params","type":"tuple"}],"name":"exactOutputSingle","outputs":[{"internalType":"uint256","name":"amountIn","type":"uint256"}],"stateMutability":"payable","type":"function"},{"inputs":[],"name":"factory","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes[]","name":"data","type":"bytes[]"}],"name":"multicall","outputs":[{"internalType":"bytes[]","name":"results","type":"bytes[]"}],"stateMutability":"payable","type":"function"},{"inputs":[],"name":"refundETH","outputs":[],"stateMutability":"payable","type":"function"},{"inputs":[{"internalType":"address","name":"token","type":"address"},{"internalType":"uint256","name":"value","type":"uint256"},{"internalType":"uint256","name":"deadline","type":"uint256"},{"internalType":"uint8","name":"v","type":"uint8"},{"internalType":"bytes32","name":"r","type":"bytes32"},{"internalType":"bytes32","name":"s","type":"bytes32"}],"name":"selfPermit","outputs":[],"stateMutability":"payable","type":"function"},{"inputs":[{"internalType":"address","name":"token","type":"address"},{"internalType":"uint256","name":"nonce","type":"uint256"},{"internalType":"uint256","name":"expiry","type":"uint256"},{"internalType":"uint8","name":"v","type":"uint8"},{"internalType":"bytes32","name":"r","type":"bytes32"},{"internalType":"bytes32","name":"s","type":"bytes32"}],"name":"selfPermitAllowed","outputs":[],"stateMutability":"payable","type":"function"},{"inputs":[{"internalType":"address","name":"token","type":"address"},{"internalType":"uint256","name":"nonce","type":"uint256"},{"internalType":"uint256","name":"expiry","type":"uint256"},{"internalType":"uint8","name":"v","type":"uint8"},{"internalType":"bytes32","name":"r","type":"bytes32"},{"internalType":"bytes32","name":"s","type":"bytes32"}],"name":"selfPermitAllowedIfNecessary","outputs":[],"stateMutability":"payable","type":"function"},{"inputs":[{"internalType":"address","name":"token","type":"address"},{"internalType":"uint256","name":"value","type":"uint256"},{"internalType":"uint256","name":"deadline","type":"uint256"},{"internalType":"uint8","name":"v","type":"uint8"},{"internalType":"bytes32","name":"r","type":"bytes32"},{"internalType":"bytes32","name":"s","type":"bytes32"}],"name":"selfPermitIfNecessary","outputs":[],"stateMutability":"payable","type":"function"},{"inputs":[{"internalType":"address","name":"token","type":"address"},{"internalType":"uint256","name":"amountMinimum","type":"uint256"},{"internalType":"address","name":"recipient","type":"address"}],"name":"sweepToken","outputs":[],"stateMutability":"payable","type":"function"},{"inputs":[{"internalType":"address","name":"token","type":"address"},{"internalType":"uint256","name":"amountMinimum","type":"uint256"},{"internalType":"address","name":"recipient","type":"address"},{"internalType":"uint256","name":"feeBips","type":"uint256"},{"internalType":"address","name":"feeRecipient","type":"address"}],"name":"sweepTokenWithFee","outputs":[],"stateMutability":"payable","type":"function"},{"inputs":[{"internalType":"int256","name":"amount0Delta","type":"int256"},{"internalType":"int256","name":"amount1Delta","type":"int256"},{"internalType":"bytes","name":"_data","type":"bytes"}],"name":"uniswapV3SwapCallback","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"uint256","name":"amountMinimum","type":"uint256"},{"internalType":"address","name":"recipient","type":"address"}],"name":"unwrapWETH9","outputs":[],"stateMutability":"payable","type":"function"},{"inputs":[{"internalType":"uint256","name":"amountMinimum","type":"uint256"},{"internalType":"address","name":"recipient","type":"address"},{"internalType":"uint256","name":"feeBips","type":"uint256"},{"internalType":"address","name":"feeRecipient","type":"address"}],"name":"unwrapWETH9WithFee","outputs":[],"stateMutability":"payable","type":"function"},{"stateMutability":"payable","type":"receive"}]'
    else:
        raise KeyError("Unknown contract name")
    """
    with open('./oaiv/tools/abidata/names.json', 'r') as file:
        names = json.load(file)
    if name in names.keys():
        file_address = names[name]
    else:
        raise KeyError("Unknown contract name")
    
    with open(file_address, 'r') as file:
        abi = json.load(file)
    """
    contract = w3.eth.contract(address=address, abi=abi)

    return contract


def find_address(name):

    if name == 'QUOTER':
        contract_address = '0xb27308f9F90D607463bb33eA1BeBb41C27CE5AB6'
    elif name == 'ROUTER':
        contract_address = '0xE592427A0AEce92De3Edee1F18E0157C05861564'
    elif name == 'ETH':
        contract_address = '0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2'
    elif name == 'USDC':
        contract_address = '0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48'
    else:
        raise KeyError("Unknown contract name")

    """
    if network in ['mainnet', 'goerli', 'ropsten']:

        with open('./oaiv/tools/address/{0}/info.json'.format(network), 'r') as file:
            names = json.load(file)
        if name in names.keys():
            contract_address = names[name]['address']
        else:
            raise KeyError("Unknown contract name")

        return contract_address

    else:

        raise KeyError("Invalid network name")
    """

    return contract_address



#
import time


#


#


#
class Uniswap:
    def __init__(
            self,
            address,
            public_key,
            private_key,
            provider,
            web3,
            router_contract_address,
            quoter_contract_address,
            router_contract,
            quoter_contract,
    ):
        self.address = address
        self.public_key = public_key
        self.private_key = private_key
        self.provider = provider
        self.web3 = web3
        self.router_contract_address = router_contract_address
        self.quoter_contract_address = quoter_contract_address

        self.router = router_contract
        self.quoter = quoter_contract

    @staticmethod
    def _deadline():
        return int(time.time()) + 10 * 60

    def make_trade_input(
            self,
            input_token_address,
            output_token_address,
            quantity_in_token,
            fee,
            slippage
    ):
        sqrtPriceLimitX96 = 0
        price = self.quoter.functions.quoteExactInputSingle(
                input_token_address, output_token_address, fee, quantity_in_token, sqrtPriceLimitX96
            ).call()
        min_tokens_bought = int(
            (1 - slippage)
            * price
        )

        contract_function = self.router.functions.exactInputSingle(
                    {
                        "tokenIn": input_token_address,
                        "tokenOut": output_token_address,
                        "fee": fee,
                        "recipient": self.address,
                        "deadline": self._deadline(),
                        "amountIn": quantity_in_token,
                        "amountOutMinimum": min_tokens_bought,
                        "sqrtPriceLimitX96": sqrtPriceLimitX96,
                    },
                )

        tx_params = {
            "from": self.public_key,
            "value": quantity_in_token,
            "nonce": self.web3.eth.get_transaction_count(self.address),
        }

        transaction = contract_function.buildTransaction(tx_params)
        # The Uniswap V3 UI uses 20% margin for transactions
        transaction["gas"] = int(self.web3.eth.estimate_gas(transaction) * 1.2)

        signed_txn = self.web3.eth.account.sign_transaction(
            transaction, private_key=self.private_key
        )

        tx_id = self.web3.toHex(self.web3.eth.send_raw_transaction(signed_txn.rawTransaction))

        return tx_id



#
import setuptools
from setuptools import setup


metadata = {'name': 'oaiv',
            'maintainer': 'Edward Azizov',
            'maintainer_email': 'edazizovv@gmail.com',
            'description': 'Blockchain Interaction Library',
            'license': 'MIT',
            'url': 'https://github.com/edazizovv/oaiv',
            'download_url': 'https://github.com/edazizovv/oaiv',
            'packages': setuptools.find_packages(),
            'include_package_data': True,
            'version': '0.4.1',
            'long_description': '',
            'python_requires': '>=3.10',
            'install_requires': []}

setup(**metadata)
